/*----- PROTECTED REGION ID(Agilent4uhv.cpp) ENABLED START -----*/
//=============================================================================
//
// file :        Agilent4uhv.cpp
//
// description : C++ source for the Agilent4uhv class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               Agilent4uhv are implemented in this file.
//
// project :     
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
// $Author: alessio $
//
// $Revision: 1.18 $
// $Date: 2018-05-10 08:04:51 $
//
// $HeadURL:  $
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================

#include <Agilent4uhv.h>
#include <Agilent4uhvClass.h>

#define STX 0x02
#define ETX 0x03
#define ADDR 0x80
#define WIN_LENGTH 3
#define CRC_LENGTH 2
#define ERROR_MESSAGE_SIZE 6
#define HEADER_SIZE 6
#define FOOTER_SIZE 3
#define NUMERIC_PAYLOAD_SIZE 6
#define ALPHANUMERIC_PAYLOAD_SIZE 10

#define MAX_DEVSTRING_LENGTH 255

/*----- PROTECTED REGION END -----*/	//	Agilent4uhv.cpp

/**
 *  Agilent4uhv class description:
 *    
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name  |  Method name
//================================================================
//  State         |  Inherited (no method)
//  Status        |  Inherited (no method)
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  Model           |  Tango::DevString	Scalar
//  SerialNumber    |  Tango::DevString	Scalar
//  TemperatureFan  |  Tango::DevDouble	Scalar
//================================================================

namespace Agilent4uhv_ns
{
/*----- PROTECTED REGION ID(Agilent4uhv::namespace_starting) ENABLED START -----*/

//	static initializations
template<typename F, typename T> void convert(const F &data, T &result)
{
	stringstream ss;
	ss << data;
	ss >> result;

	if (ss.fail()) {
		stringstream err_mesg;
		err_mesg << "Conversion fail of " << data;
		Tango::Except::throw_exception( "",
				err_mesg.str(),
				"convert()");
	}
}

template<typename F, typename T> void convert(const F &data, double &result)
{
	stringstream ss;
	ss << data;

	if (ss.str() == "Inf")
		result = INFINITY;
	else
		ss >> result;

	if (ss.fail()) {
		stringstream err_mesg;
		err_mesg << "Conversion fail of " << data;
		Tango::Except::throw_exception( "",
				err_mesg.str(),
				"convert()");
	}
}

template<typename F> void convert(const F &data, string &result)
{
	stringstream ss;
	ss << fixed;
	ss << data;
	result = ss.str();
	if (ss.fail()) {
		stringstream err_mesg;
		err_mesg << "Conversion fail of " << data;
		Tango::Except::throw_exception( "",
				err_mesg.str(),
				"convert()");
	}
}

/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : Agilent4uhv::Agilent4uhv()
 *	Description : Constructors for a Tango device
 *                implementing the classAgilent4uhv
 */
//--------------------------------------------------------
Agilent4uhv::Agilent4uhv(Tango::DeviceClass *cl, string &s)
 : Serial2Client(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(Agilent4uhv::constructor_1) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::constructor_1
}
//--------------------------------------------------------
Agilent4uhv::Agilent4uhv(Tango::DeviceClass *cl, const char *s)
 : Serial2Client(cl, s)
{
	/*----- PROTECTED REGION ID(Agilent4uhv::constructor_2) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::constructor_2
}
//--------------------------------------------------------
Agilent4uhv::Agilent4uhv(Tango::DeviceClass *cl, const char *s, const char *d)
 : Serial2Client(cl, s, d)
{
	/*----- PROTECTED REGION ID(Agilent4uhv::constructor_3) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::constructor_3
}

//--------------------------------------------------------
/**
 *	Method      : Agilent4uhv::delete_device()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void Agilent4uhv::delete_device()
{
	DEBUG_STREAM << "Agilent4uhv::delete_device() " << device_name << endl;
	/*----- PROTECTED REGION ID(Agilent4uhv::delete_device) ENABLED START -----*/
	
	//	Delete device allocated objects
	if (worker) {
		worker->stop();
		worker->join(NULL);
	}

	//	Delete device allocated objects
	if (attr_Model_read[0])
		delete attr_Model_read[0];
	if (attr_SerialNumber_read[0])
		delete attr_SerialNumber_read[0];

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::delete_device
	delete[] attr_Model_read;
	delete[] attr_SerialNumber_read;
	delete[] attr_TemperatureFan_read;

	if (Tango::Util::instance()->is_svr_shutting_down()==false  &&
		Tango::Util::instance()->is_device_restarting(device_name)==false &&
		Tango::Util::instance()->is_svr_starting()==false)
	{
		//	If not shutting down call delete device for inherited object
		Serial2Client_ns::Serial2Client::delete_device();
	}
}

//--------------------------------------------------------
/**
 *	Method      : Agilent4uhv::init_device()
 *	Description : will be called at device initialization.
 */
//--------------------------------------------------------
void Agilent4uhv::init_device()
{
	DEBUG_STREAM << "Agilent4uhv::init_device() create device " << device_name << endl;
	/*----- PROTECTED REGION ID(Agilent4uhv::init_device_before) ENABLED START -----*/
	
	//	Initialization before get_device_property() call
	mandatoryNotDefined = false;
	channels.clear();
	init_error.clear();
	worker = NULL;

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::init_device_before
	
	if (Tango::Util::instance()->is_svr_starting() == false  &&
		Tango::Util::instance()->is_device_restarting(device_name)==false)
	{
		//	If not starting up call init device for inherited object
		Serial2Client_ns::Serial2Client::init_device();
	}

	//	Get the device properties from database
	get_device_property();
	
	attr_Model_read = new Tango::DevString[1];
	attr_SerialNumber_read = new Tango::DevString[1];
	attr_TemperatureFan_read = new Tango::DevDouble[1];
	//	No longer if mandatory property not set. 
	if (mandatoryNotDefined)
		return;

	/*----- PROTECTED REGION ID(Agilent4uhv::init_device) ENABLED START -----*/

	//	Initialize device
	try {
		attr_Model_read[0] = new char[MAX_DEVSTRING_LENGTH];
		attr_SerialNumber_read[0] = new char[MAX_DEVSTRING_LENGTH];

		/* Che if the mBar is pressure unit used */
		int pressure_unit;
		read_window(600, pressure_unit);
		if (pressure_unit != 1) {
			Tango::Except::throw_exception( "",
					"mBar is not the pressure unit used",
					"Agilent4uhv::init_device()");
		}

		/* Identify model:
		* 9299200/1 con 2 canali x 80W  --> attivi canali 1 e 2
		* 9299020   con 2 canali x 200W --> attivi canali 1 e 3
		* 9299400   con 4 canali x 80W  --> attivi canali da 1 a 4
		*/
		string model;
		read_window(319, model);
		if (model == "9299200" || model == "9299201") {
			channels.push_back("1");
			channels.push_back("2");
		} else if (model == "9299020") {
			channels.push_back("1");
			channels.push_back("3");
		} else if (model == "9299400" || model == "9299401") {
			channels.push_back("1");
			channels.push_back("2");
			channels.push_back("3");
			channels.push_back("4");
		} else {
			Tango::Except::throw_exception( "",
					"Unknown model",
					"Agilent4uhv::init_device()");
		}
		strncpy(attr_Model_read[0], model.c_str(), MAX_DEVSTRING_LENGTH);

		// Retrieve serial number
		string serial;
		read_window(323, serial);
		strncpy(attr_SerialNumber_read[0], serial.c_str(), MAX_DEVSTRING_LENGTH);

		/* Read all the device data */
		static const int nw[] = { 8, /*, 108, 205, 320, 319, 323, 503, 504, 600 */
			601, 602, 603, 800, 801, 802, 803, 804, 808, 809, 810, 811};
		vector<int> normal_windows(nw, nw + sizeof(nw) / sizeof(nw[0]));
		for (size_t i=0; i<channels.size(); ++i) {
			int chan;
			convert(channels[i], chan);
			normal_windows.push_back(10 + chan);
			normal_windows.push_back(2060 + chan);
			normal_windows.push_back(600 + chan * 10);
			normal_windows.push_back(602 + chan * 10);
			normal_windows.push_back(603 + chan * 10);
			normal_windows.push_back(604 + chan * 10);
			normal_windows.push_back(605 + chan * 10);
			normal_windows.push_back(800 + chan * 10);
			normal_windows.push_back(801 + chan * 10);
		}
		for (size_t i=0; i<normal_windows.size(); ++i) {
			read_window_into_cache(normal_windows[i]);
		}
		static const int iw[] = { 812, 822, 832, 842 };
		vector<int> important_windows(iw, iw + sizeof(iw) / sizeof(iw[0]));
		for (size_t i=0; i<important_windows.size(); ++i) {
			read_window_into_cache(important_windows[i]);
		}
		read_window_into_cache(205);

		// Contruct polled windows vector
		for (size_t i=0; i<normal_windows.size(); ++i) {
			polled_windows.push_back(normal_windows[i]);
			polled_windows.insert(polled_windows.end(),
					important_windows.begin(),
					important_windows.end());
		}
		polled_windows.push_back(205 /* it MUST be latest */);

		check_state();

		/* Start the poller */
		worker = new Agilent4uhvWorker(this);
		worker->start();
	} catch(Tango::DevFailed &e) {
		if (string(e.errors[0].reason) == "API_CommandNotAllowed")
			init_error = "Initialization failed: " + get_status2();
		else
			init_error = "Initialization failed: " + string(e.errors[0].desc);
	} catch (...) {
		init_error = "Initialization failed: Unknown error";
	}

	if (! init_error.empty()) {
		ERROR_STREAM << init_error << endl;
		set_state(Tango::UNKNOWN);
		set_status(init_error);
		assert(false);
	}

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::init_device
}

//--------------------------------------------------------
/**
 *	Method      : Agilent4uhv::get_device_property()
 *	Description : Read database to initialize property data members.
 */
//--------------------------------------------------------
void Agilent4uhv::get_device_property()
{
	/*----- PROTECTED REGION ID(Agilent4uhv::get_device_property_before) ENABLED START -----*/
	
	//	Initialize property data members
	
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::get_device_property_before

	mandatoryNotDefined = false;

	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("DeviceName"));
	dev_prop.push_back(Tango::DbDatum("PollingSleep"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);
	
		//	get instance on Agilent4uhvClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		Agilent4uhvClass	*ds_class =
			(static_cast<Agilent4uhvClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize DeviceName from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  deviceName;
		else {
			//	Try to initialize DeviceName from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  deviceName;
		}
		//	And try to extract DeviceName value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  deviceName;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize PollingSleep from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  pollingSleep;
		else {
			//	Try to initialize PollingSleep from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  pollingSleep;
		}
		//	And try to extract PollingSleep value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  pollingSleep;

	}

	/*----- PROTECTED REGION ID(Agilent4uhv::get_device_property_after) ENABLED START -----*/
	
	//	Check device property data members init
	if (pollingSleep > 1000000) {
		pollingSleep = 1000000;
	}

	if (pollingSleep < 0) {
		pollingSleep = 0;
	}
	
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::get_device_property_after
}
//--------------------------------------------------------
/**
 *	Method      : Agilent4uhv::check_mandatory_property()
 *	Description : For mandatory properties check if defined in database.
 */
//--------------------------------------------------------
void Agilent4uhv::check_mandatory_property(Tango::DbDatum &class_prop, Tango::DbDatum &dev_prop)
{
	//	Check if all properties are empty
	if (class_prop.is_empty() && dev_prop.is_empty())
	{
		TangoSys_OMemStream	tms;
		tms << endl <<"Property \'" << dev_prop.name;
		if (Tango::Util::instance()->_UseDb==true)
			tms << "\' is mandatory but not defined in database";
		else
			tms << "\' is mandatory but cannot be defined without database";
		string	status(get_status());
		status += tms.str();
		set_status(status);
		mandatoryNotDefined = true;
		/*----- PROTECTED REGION ID(Agilent4uhv::check_mandatory_property) ENABLED START -----*/
		cerr << tms.str() << " for " << device_name << endl;
		
		/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::check_mandatory_property
	}
}


//--------------------------------------------------------
/**
 *	Method      : Agilent4uhv::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void Agilent4uhv::always_executed_hook()
{
	DEBUG_STREAM << "Agilent4uhv::always_executed_hook()  " << device_name << endl;
	if (mandatoryNotDefined)
	{
		string	status(get_status());
		Tango::Except::throw_exception(
					(const char *)"PROPERTY_NOT_SET",
					status.c_str(),
					(const char *)"Agilent4uhv::always_executed_hook()");
	}
	/*----- PROTECTED REGION ID(Agilent4uhv::always_executed_hook) ENABLED START -----*/
	
	//	code always executed before all requests
	if (! init_error.empty()) {
		ERROR_STREAM << init_error << endl;
		set_status(init_error);
		set_state(Tango::UNKNOWN);
		assert(false);
		return;
	}

	try {
		check_state();
	} catch(Tango::DevFailed &e) {
		ERROR_STREAM << string(e.errors[0].desc) << endl;
		set_state(Tango::UNKNOWN);
		set_status(string(e.errors[0].desc));
		assert(false);
	}

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : Agilent4uhv::read_attr_hardware()
 *	Description : Hardware acquisition for attributes
 */
//--------------------------------------------------------
void Agilent4uhv::read_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "Agilent4uhv::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(Agilent4uhv::read_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	check_init();
	
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute Model related method
 *	Description: 
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_Model(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_Model(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(Agilent4uhv::read_Model) ENABLED START -----*/
	attr.set_value(attr_Model_read);
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_Model
}
//--------------------------------------------------------
/**
 *	Read attribute SerialNumber related method
 *	Description: 
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_SerialNumber(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_SerialNumber(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(Agilent4uhv::read_SerialNumber) ENABLED START -----*/
	attr.set_value(attr_SerialNumber_read);
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_SerialNumber
}
//--------------------------------------------------------
/**
 *	Read attribute TemperatureFan related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_TemperatureFan(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_TemperatureFan(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(Agilent4uhv::read_TemperatureFan) ENABLED START -----*/
	//	Set the attribute value
	int value;
	read_window_from_cache(800, value);
	convert(value, attr_TemperatureFan_read[0]);
	attr.set_value(attr_TemperatureFan_read);
	
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_TemperatureFan
}

//--------------------------------------------------------
/**
 *	Read attribute HVX related method
 *	Description: 
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_HVX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_HVX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevState	*att_value = get_HVX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_HVX) ENABLED START -----*/
	int win;
	convert("01" + attr.get_name().substr(attr.get_name().size()-1), win);	

	bool value;
	read_window_from_cache(win, value);
	if (value)
		*att_value = Tango::ON;
	else
		*att_value = Tango::OFF;
	attr.set_value(att_value);
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_HVX
}
//--------------------------------------------------------
/**
 *	Read attribute DeviceNumberX related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_DeviceNumberX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_DeviceNumberX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevLong	*att_value = get_DeviceNumberX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_DeviceNumberX) ENABLED START -----*/
	int win;
	convert("6" + attr.get_name().substr(attr.get_name().size()-1) + "0", win);	

	int value;
	read_window_from_cache(win, value);
	*att_value = value;
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_DeviceNumberX
}
//--------------------------------------------------------
/**
 *	Write attribute DeviceNumberX related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::write_DeviceNumberX(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::write_DeviceNumberX(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(Agilent4uhv::write_DeviceNumberX) ENABLED START -----*/
	check_init();

	int win;
	convert("6" + attr.get_name().substr(attr.get_name().size()-1) + "0", win);	

	int value = w_val;
	write_window(win, value);
	read_window_into_cache(win);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::write_DeviceNumberX
}
//--------------------------------------------------------
/**
 *	Read attribute PowerMaxX related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_PowerMaxX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_PowerMaxX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevLong	*att_value = get_PowerMaxX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_PowerMaxX) ENABLED START -----*/
	int win;
	convert("6" + attr.get_name().substr(attr.get_name().size()-1) + "2", win);	

	int value;
	read_window_from_cache(win, value);
	*att_value = value;
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_PowerMaxX
}
//--------------------------------------------------------
/**
 *	Write attribute PowerMaxX related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::write_PowerMaxX(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::write_PowerMaxX(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(Agilent4uhv::write_PowerMaxX) ENABLED START -----*/
	check_init();

	int win;
	convert("6" + attr.get_name().substr(attr.get_name().size()-1) + "2", win);

	int value = w_val;
	write_window(win, value);
	read_window_into_cache(win);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::write_PowerMaxX
}
//--------------------------------------------------------
/**
 *	Read attribute VTargetX related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_VTargetX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_VTargetX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevLong	*att_value = get_VTargetX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_VTargetX) ENABLED START -----*/
	int win;
	convert("6" + attr.get_name().substr(attr.get_name().size()-1) + "3", win);

	int value;
	read_window_from_cache(win, value);
	*att_value = value;
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_VTargetX
}
//--------------------------------------------------------
/**
 *	Write attribute VTargetX related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::write_VTargetX(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::write_VTargetX(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(Agilent4uhv::write_VTargetX) ENABLED START -----*/
	check_init();

	int win;
	convert("6" + attr.get_name().substr(attr.get_name().size()-1) + "3", win);

	int value = w_val;
	write_window(win, value);
	read_window_into_cache(win);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::write_VTargetX
}
//--------------------------------------------------------
/**
 *	Read attribute IProtectX related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_IProtectX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_IProtectX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevLong	*att_value = get_IProtectX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_IProtectX) ENABLED START -----*/
	int win;
	convert("6" + attr.get_name().substr(attr.get_name().size()-1) + "4", win);

	int value;
	read_window_from_cache(win, value);
	*att_value = value;
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_IProtectX
}
//--------------------------------------------------------
/**
 *	Write attribute IProtectX related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::write_IProtectX(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::write_IProtectX(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(Agilent4uhv::write_IProtectX) ENABLED START -----*/
	check_init();

	int win;
	convert("6" + attr.get_name().substr(attr.get_name().size()-1) + "4", win);

	int value = w_val;
	write_window(win, value);
	read_window_into_cache(win);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::write_IProtectX
}
//--------------------------------------------------------
/**
 *	Read attribute SetPointX related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_SetPointX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_SetPointX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevDouble	*att_value = get_SetPointX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_SetPointX) ENABLED START -----*/
	int win;
	convert("6" + attr.get_name().substr(attr.get_name().size()-1) + "5", win);
	
	read_window_from_cache(win, *att_value);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_SetPointX
}
//--------------------------------------------------------
/**
 *	Write attribute SetPointX related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::write_SetPointX(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::write_SetPointX(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(Agilent4uhv::write_SetPointX) ENABLED START -----*/
	int win;
	convert("6" + attr.get_name().substr(attr.get_name().size()-1) + "5", win);

	write_window(win, w_val);
	read_window_into_cache(win);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::write_SetPointX
}
//--------------------------------------------------------
/**
 *	Read attribute TemperatureHVX related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_TemperatureHVX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_TemperatureHVX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevDouble	*att_value = get_TemperatureHVX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_TemperatureHVX) ENABLED START -----*/
	int channel, win;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);
	switch(channel) {
		case 1:	win = 801; break;
		case 2:	win = 802; break;
		case 3: win = 808; break;
		case 4: win = 809; break;
		default:
			Tango::Except::throw_exception( "",
					"This channel doesn't exist",
					"Agilent4uhv::read_TemperatureHVX()");
	}
	int value;
	read_window_from_cache(win, value);
	convert(value, *att_value);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_TemperatureHVX
}
//--------------------------------------------------------
/**
 *	Read attribute VoltageX related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_VoltageX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_VoltageX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevDouble	*att_value = get_VoltageX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_VoltageX) ENABLED START -----*/
	int win;
	convert("8" + attr.get_name().substr(attr.get_name().size()-1) + "0", win);

	int value;
	read_window_from_cache(win, value);
	convert(value, *att_value);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_VoltageX
}
//--------------------------------------------------------
/**
 *	Read attribute CurrentX related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_CurrentX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_CurrentX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevDouble	*att_value = get_CurrentX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_CurrentX) ENABLED START -----*/
	int win;
	convert("8" + attr.get_name().substr(attr.get_name().size()-1) + "1", win);
	read_window_from_cache(win, *att_value);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_CurrentX
}
//--------------------------------------------------------
/**
 *	Read attribute PressureX related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_PressureX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_PressureX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevDouble	*att_value = get_PressureX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_PressureX) ENABLED START -----*/
	int win;
	convert("8" + attr.get_name().substr(attr.get_name().size()-1) + "2", win);
	read_window_from_cache(win, *att_value);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_PressureX
}
//--------------------------------------------------------
/**
 *	Read attribute StateX related method
 *	Description: 
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_StateX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_StateX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevState	*att_value = get_StateX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_StateX) ENABLED START -----*/
	//	Set the attribute value
	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

	int errorcodes;
	read_window_from_cache(2060 + channel, errorcodes);

	bool state;
	read_window_from_cache(10 + channel, state);

	if (errorcodes) {
		*att_value = Tango::FAULT;
	} else {
		if (state)
			*att_value = Tango::ON;
		else
			*att_value = Tango::OFF;
	}

	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_StateX
}
//--------------------------------------------------------
/**
 *	Read attribute StatusX related method
 *	Description: 
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_StatusX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_StatusX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevString	*att_value = get_StatusX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_StatusX) ENABLED START -----*/
	//	Set the attribute value
	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

	string status;

	bool state;
	read_window_from_cache(10 + channel, state);
	if (state)
		status = "The device is in ON state";
	else
		status = "The device is in OFF state";

	int errorcodes;
	read_window_from_cache(2060 + channel, errorcodes);
	if (errorcodes != 0) {
		status = "";
		if (errorcodes & 0x1)
			status += "FAN error ";
		if (errorcodes & 0x2)
			status += "Power input error from HV ";
		if (errorcodes & 0x4)
			status += "Power input error from PFC ";
		if (errorcodes & 0x8)
			status += "Overtemp PFC ";
		if (errorcodes & 0x10)
			status += "Error communication CPU-HV ";
		if (errorcodes & 0x20)
			status += "Interlock cable ";
		if (errorcodes & 0x40)
			status += "Overtemp HV ";
		if (errorcodes & 0x80)
			status += "Protect ";
		if (errorcodes & 0x100)
			status += "Measurement error ";
		if (errorcodes & 0x200)
			status += "HV out error ";
		if (errorcodes & 0x400)
			status += "Short circuit ";
		if (errorcodes & 0x800)
			status += "HV disabled ";
	}
	
	strncpy(*att_value, status.c_str(), MAX_DEVSTRING_LENGTH);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_StatusX
}
//--------------------------------------------------------
/**
 *	Read attribute AutoStartX related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_AutoStartX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_AutoStartX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevBoolean	*att_value = get_AutoStartX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_AutoStartX) ENABLED START -----*/
	//	Set the attribute value
	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

 	string autostart;
 	read_window_from_cache(601, autostart);

	convert(autostart[autostart.size()-channel], *att_value);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_AutoStartX
}
//--------------------------------------------------------
/**
 *	Write attribute AutoStartX related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::write_AutoStartX(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::write_AutoStartX(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(Agilent4uhv::write_AutoStartX) ENABLED START -----*/
	check_init();

	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

 	string autostart;
 	read_window_from_cache(601, autostart);

	if (w_val)
		autostart[autostart.size()-channel] = '1';
	else
		autostart[autostart.size()-channel] = '0';

	write_window(601, autostart);
	read_window_into_cache(601);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::write_AutoStartX
}
//--------------------------------------------------------
/**
 *	Read attribute ProtectX related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_ProtectX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_ProtectX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevBoolean	*att_value = get_ProtectX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_ProtectX) ENABLED START -----*/
	//	Set the attribute value
	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

 	string protect;
 	read_window_from_cache(602, protect);

	convert(protect[protect.size()-channel], *att_value);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_ProtectX
}
//--------------------------------------------------------
/**
 *	Write attribute ProtectX related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::write_ProtectX(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::write_ProtectX(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(Agilent4uhv::write_ProtectX) ENABLED START -----*/
	check_init();

	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

 	string protect;
 	read_window_from_cache(602, protect);

	if (w_val)
		protect[protect.size()-channel] = '1';
	else
		protect[protect.size()-channel] = '0';

	write_window(602, protect);
	read_window_into_cache(602);
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::write_ProtectX
}
//--------------------------------------------------------
/**
 *	Read attribute FixedStepX related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_FixedStepX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_FixedStepX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevBoolean	*att_value = get_FixedStepX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_FixedStepX) ENABLED START -----*/
	//	Set the attribute value
	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

 	string fs;
 	read_window_from_cache(603, fs);

	convert(fs[fs.size()-channel], *att_value);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_FixedStepX
}
//--------------------------------------------------------
/**
 *	Write attribute FixedStepX related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::write_FixedStepX(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::write_FixedStepX(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(Agilent4uhv::write_FixedStepX) ENABLED START -----*/
	check_init();

	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

 	string fs;
 	read_window_from_cache(603, fs);

	if (w_val)
		fs[fs.size()-channel] = '1';
	else
		fs[fs.size()-channel] = '0';

	write_window(603, fs);
	read_window_into_cache(603);
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::write_FixedStepX
}
//--------------------------------------------------------
/**
 *	Read attribute StatusSetPointX related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_StatusSetPointX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_StatusSetPointX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevBoolean	*att_value = get_StatusSetPointX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_StatusSetPointX) ENABLED START -----*/
	//	Set the attribute value
	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

 	string ss;
 	read_window_from_cache(804, ss);

	convert(ss[ss.size()-channel], *att_value);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_StatusSetPointX
}
//--------------------------------------------------------
/**
 *	Read attribute InterlockStatusX related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_InterlockStatusX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_InterlockStatusX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevBoolean	*att_value = get_InterlockStatusX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_InterlockStatusX) ENABLED START -----*/
	//	Set the attribute value
	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

 	string interlock;
 	read_window_from_cache(803, interlock);

	convert(interlock[interlock.size()-channel], *att_value);
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_InterlockStatusX
}
//--------------------------------------------------------
/**
 *	Read attribute StatMaskX related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void Agilent4uhv::read_StatMaskX(Tango::Attribute &attr)
{
	DEBUG_STREAM << "Agilent4uhv::read_StatMaskX(Tango::Attribute &attr) entering... " << endl;
	Tango::DevLong	*att_value = get_StatMaskX_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(Agilent4uhv::read_StatMaskX) ENABLED START -----*/
	//	Set the attribute value
	int channel;
	convert(attr.get_name().substr(attr.get_name().size()-1), channel);

	*att_value = 0;

	bool state;
	read_window_from_cache(10 + channel, state);
	if (state)
		*att_value |= 0x1;

	int mode;
	read_window_from_cache(8, mode);
	if (mode != 3 /* LOCAL */)
		*att_value |= 0x1 << 1;

	int errorcodes;
	read_window_from_cache(2060 + channel, errorcodes);
	if (errorcodes != 0)
		*att_value |= 0x1 << 2;

	string fs;
	read_window_from_cache(603, fs);
 	if (fs[fs.size()-channel] == '1')
		*att_value |= 0x1 << 3;

	// FIXME bit 4,5,6 and 7 aren't implemented
	
	attr.set_value(att_value);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::read_StatMaskX
}
//--------------------------------------------------------
/**
 *	Method      : Agilent4uhv::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void Agilent4uhv::add_dynamic_attributes()
{
	//	Example to add dynamic attribute:
	//	Copy inside the following protected area to create instance(s) at startup.
	//	add_HVX_dynamic_attribute("MyHVXAttribute");
	//	add_DeviceNumberX_dynamic_attribute("MyDeviceNumberXAttribute");
	//	add_PowerMaxX_dynamic_attribute("MyPowerMaxXAttribute");
	//	add_VTargetX_dynamic_attribute("MyVTargetXAttribute");
	//	add_IProtectX_dynamic_attribute("MyIProtectXAttribute");
	//	add_SetPointX_dynamic_attribute("MySetPointXAttribute");
	//	add_TemperatureHVX_dynamic_attribute("MyTemperatureHVXAttribute");
	//	add_VoltageX_dynamic_attribute("MyVoltageXAttribute");
	//	add_CurrentX_dynamic_attribute("MyCurrentXAttribute");
	//	add_PressureX_dynamic_attribute("MyPressureXAttribute");
	//	add_StateX_dynamic_attribute("MyStateXAttribute");
	//	add_StatusX_dynamic_attribute("MyStatusXAttribute");
	//	add_AutoStartX_dynamic_attribute("MyAutoStartXAttribute");
	//	add_ProtectX_dynamic_attribute("MyProtectXAttribute");
	//	add_FixedStepX_dynamic_attribute("MyFixedStepXAttribute");
	//	add_StatusSetPointX_dynamic_attribute("MyStatusSetPointXAttribute");
	//	add_InterlockStatusX_dynamic_attribute("MyInterlockStatusXAttribute");
	//	add_StatMaskX_dynamic_attribute("MyStatMaskXAttribute");
	
	/*----- PROTECTED REGION ID(Agilent4uhv::add_dynamic_attributes) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic attributes if any
	for (size_t i=0; i<channels.size(); ++i) {
		add_HVX_dynamic_attribute("HV" + channels[i]);
		add_DeviceNumberX_dynamic_attribute("DeviceNumber" + channels[i]);
		add_PowerMaxX_dynamic_attribute("PowerMax" + channels[i]);
		add_VTargetX_dynamic_attribute("VTarget" + channels[i]);
		add_IProtectX_dynamic_attribute("IProtect" + channels[i]);
		add_SetPointX_dynamic_attribute("SetPoint" + channels[i]);
		add_TemperatureHVX_dynamic_attribute("TemperatureHV" + channels[i]);
		add_VoltageX_dynamic_attribute("Voltage" + channels[i]);
		add_CurrentX_dynamic_attribute("Current" + channels[i]);
		add_PressureX_dynamic_attribute("Pressure" + channels[i]);
		add_StateX_dynamic_attribute("State" + channels[i]);

		string attr_status_name = "Status" + channels[i];
		add_StatusX_dynamic_attribute(attr_status_name);
		Tango::DevString *att_value = get_StatusX_data_ptr(attr_status_name);
		*att_value = new char[MAX_DEVSTRING_LENGTH];

		add_AutoStartX_dynamic_attribute("AutoStart" + channels[i]);
		add_ProtectX_dynamic_attribute("Protect" + channels[i]);
		add_FixedStepX_dynamic_attribute("Fixed" + channels[i]);
		add_StatusSetPointX_dynamic_attribute("StatusSetPoint" + channels[i]);
		add_InterlockStatusX_dynamic_attribute("InterlockStatus" + channels[i]);
		add_StatMaskX_dynamic_attribute("StatMask" + channels[i]);
	}
	
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command OffHVX related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void Agilent4uhv::off_hvx(Tango::Command &command)
{
	DEBUG_STREAM << "Agilent4uhv::" << command.get_name() << "  - " << device_name << endl;
	/*----- PROTECTED REGION ID(Agilent4uhv::off_hvx) ENABLED START -----*/
	
	//	Add your own code
 	check_init();
	
	int win;
	convert("01" + command.get_name().substr(command.get_name().size()-1), win);	

	write_window(win, false);
	read_window_into_cache(win);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::off_hvx
}
//--------------------------------------------------------
/**
 *	Command OnHVX related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void Agilent4uhv::on_hvx(Tango::Command &command)
{
	DEBUG_STREAM << "Agilent4uhv::" << command.get_name() << "  - " << device_name << endl;
	/*----- PROTECTED REGION ID(Agilent4uhv::on_hvx) ENABLED START -----*/
	
	//	Add your own code
 	check_init();
	
	int win;
	convert("01" + command.get_name().substr(command.get_name().size()-1), win);	

	write_window(win, true);
	read_window_into_cache(win);

	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::on_hvx
}
//--------------------------------------------------------
/**
 *	Method      : Agilent4uhv::add_dynamic_commands()
 *	Description : Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void Agilent4uhv::add_dynamic_commands()
{
	//	Example to add dynamic command:
	//	Copy inside the folowing protected area to instanciate at startup.
	//	add_OffHVX_dynamic_command("MyOffHVXCommand", true);
	//	add_OnHVX_dynamic_command("MyOnHVXCommand", true);
	
	/*----- PROTECTED REGION ID(Agilent4uhv::add_dynamic_commands) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic commands if any
	for (size_t i=0; i<channels.size(); ++i) {
		add_OnHVX_dynamic_command("OnHV" + channels[i], true);
		add_OffHVX_dynamic_command("OffHV" + channels[i], true);
	}
	
	/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::add_dynamic_commands
}

/*----- PROTECTED REGION ID(Agilent4uhv::namespace_ending) ENABLED START -----*/

//	Additional Methods
void Agilent4uhv::check_init()
{
	if (! init_error.empty())
		Tango::Except::throw_exception( "",
				init_error.c_str(),
				"Agilent4uhv::check_init()");
}

void Agilent4uhv::init_header(const int win, const windowType type, 
		vector<unsigned char> &header)
{
	header.push_back(STX);
	header.push_back(ADDR);

	char winc[WIN_LENGTH+1];
	snprintf(winc, WIN_LENGTH+1, "%03d", win);
	for (size_t i=0; i<WIN_LENGTH; ++i)
		header.insert(header.end(), winc[i]);

	header.insert(header.end(), type);
}

int Agilent4uhv::calc_crc(const vector<unsigned char> &header, 
		const vector<unsigned char> &payload) {
	int crc = 0;
	for (size_t i=1 /* Skip STX */; i < header.size(); ++i)
		crc ^= header[i];
	for (size_t i=0; i < payload.size(); ++i)
		crc ^= payload[i];
	crc ^= ETX;
	return crc;
}

void Agilent4uhv::check_crc(const vector<unsigned char> &header, 
		const vector<unsigned char> &payload, const vector<unsigned char> &footer) {
	char crc[CRC_LENGTH+1];
	snprintf(crc, CRC_LENGTH+1, "%02X", calc_crc(header, payload));
	if(memcmp(footer.data()+1 /* Skip ETX */, crc, CRC_LENGTH) != 0)
		Tango::Except::throw_exception( "",
				"CRC check failed",
				"Message::check_crc()");
}

void Agilent4uhv::init_footer(const int computed_crc, vector<unsigned char> &footer)
{
	char crc[CRC_LENGTH+1];
	snprintf(crc, CRC_LENGTH+1, "%02X", computed_crc);
	
	footer.push_back(ETX);
	footer.push_back(0);
	footer.push_back(0);
	memcpy(footer.data()+1 /* Skip ETX */, crc, CRC_LENGTH);
}

/* Must be called with iomutex held */
void Agilent4uhv::send_message(const int win, const windowType type, 
		const vector<unsigned char> &payload)
{
	vector<unsigned char> header, footer;
	int attempts = 2;

	init_header(win, type, header);
	int crc = calc_crc(header, payload);
	init_footer(crc, footer);

	vector<unsigned char> message;
	message.insert(message.end(), header.begin(), header.end());
	message.insert(message.end(), payload.begin(), payload.end());
	message.insert(message.end(), footer.begin(), footer.end());

retry:
	try {
		Serial2Client::write(message);
	} catch (Tango::DevFailed &e) {
		if (--attempts) {
			// Waiting the device
			int retries = 0;
			while(get_state2() != Tango::ON && retries++ < 5) {
				sleep(1);
			}
			goto retry;
		}
		throw;
	}
	assert(Serial2Client::output_length() == 0);
}

/* Must be called with iomutex held */
void Agilent4uhv::receive_message(vector<unsigned char> &payload)
{
	vector<unsigned char> message, crc;
	int attempts = 5;

retry:
	try {
		Serial2Client::read_until(ETX, message);
		Serial2Client::read(CRC_LENGTH, crc);
	} catch (Tango::DevFailed &e) {
		if (--attempts) {
			goto retry;
		}
		throw;
	}
	assert(Serial2Client::input_length() == 0);

	message.insert(message.end(), crc.begin(), crc.end());

	/* Error message is fixed-size */
	if (message.size() == ERROR_MESSAGE_SIZE) {
		string err_mesg;
		switch (message[2]) {
			case 0x06: /* ACK */
				break;
			case 0x15: /* NACK */
				err_mesg = "Execution of the command has failed";
				break;
			case 0x32: /* Unknown Window */
				err_mesg = "The window specified in the command"
					" is not a valid window";
				break;
			case 0x33: /* Data Type Error */
				err_mesg = "The data type specified in the command"
					" (Logic, Numeric or Alphanumeric) is not"
					" in agreement with the Window specified";
				break;
			case 0x35: /* Win Disabled */
				err_mesg = "The window specified is Read Only or"
					" is temporarily disabled";
				break;
			default:
				err_mesg = "The value expressed during a write"
					" command is not within the range value"
					" for the specified window.";
		}
		if (! err_mesg.empty())
			Tango::Except::throw_exception( "",
					err_mesg,
					"Agilent4uhv::receive_message()");

	} else {
		size_t payload_size = message.size() - (HEADER_SIZE + FOOTER_SIZE);
		payload.assign(message.begin() + HEADER_SIZE,
				message.begin() + HEADER_SIZE + payload_size);
	}
}

/* Must be called with iomutex held */
void Agilent4uhv::_send_receive_message(const int win, const windowType type, 
		const vector<unsigned char> &answer, vector<unsigned char> &reply)
{
	send_message(win, type, answer);
	receive_message(reply);
}

void Agilent4uhv::send_receive_message(const int win, const windowType type, 
		const vector<unsigned char> &answer, vector<unsigned char> &reply)
{
	omni_mutex_lock guardio(iomutex);
	_send_receive_message(win, type, answer, reply);
}

void Agilent4uhv::decode(const vector<unsigned char> &payload, bool &data)
{
	convert(payload[0], data);
}

void Agilent4uhv::encode(const bool &data, vector<unsigned char> &payload)
{
	payload.resize(1);
	convert(data, payload[0]);
}

void Agilent4uhv::decode(const vector<unsigned char> &payload, int &data)
{
	string s(payload.begin(), payload.end());
	convert(s, data);
}

void Agilent4uhv::encode(const int &data, vector<unsigned char> &payload)
{
	string s;
	convert(data, s);
	/* Numeric is exactly 6 characters right-justified and 
	 * padded with ascii zero (0x30) */
	s.insert(s.begin(), NUMERIC_PAYLOAD_SIZE - s.size(), '0');
	payload.assign(s.begin(), s.end());
}

/* Alphanumeric, which is 10 characters long, when used as double (scientific notation) 
 * is righ-justified and padded with spaces (0x20) */
void Agilent4uhv::decode(const vector<unsigned char> &payload, double &data)
{
	// Left trim of spaces
	size_t start;
	for (size_t i=0; i<payload.size(); ++i) {
		if (payload[i] != ' ') {
			start = i;
			break;
		}
	}
	string s = string(payload.begin() + start, payload.end());
	convert(s, data);
}

void Agilent4uhv::encode(const double &data, vector<unsigned char> &payload)
{
	char text[ALPHANUMERIC_PAYLOAD_SIZE + 1];
	snprintf(text, ALPHANUMERIC_PAYLOAD_SIZE + 1, "%.1E", data);
	string s(text);
	s.insert(s.begin(), ALPHANUMERIC_PAYLOAD_SIZE - s.size(), ' ');
	payload.assign(s.begin(), s.end());
}

/* Alphanumeric, which is 10 characters long, when used as string
 * is left-justified and padded with spaces (0x20). When used as decimal number is
 * righ-justified and padded with zeros (x30). In this last case the code should
 * explicitly define 10 charaters to avoid automatic padding and pad with zeros
 * directly */
void Agilent4uhv::decode(const vector<unsigned char> &payload, string &data)
{
	// Right trim of spaces
	size_t end;
	for (size_t i=payload.size(); i>0; --i) {
		if (payload[i-1] != ' ') {
			end = i;
			break;
		}
	}
	data = string(payload.begin(), payload.begin() + end);
}

void Agilent4uhv::encode(const string &data, vector<unsigned char> &payload)
{
	string s = data;
	s.insert(s.end(), ALPHANUMERIC_PAYLOAD_SIZE - s.size(), ' ');
	payload.assign(s.begin(), s.end());
}

template<typename T> void Agilent4uhv::read_window(const int win, T &data)
{
	vector<unsigned char> discard, payload;
	send_receive_message(win, READ_WINDOW, discard, payload);
	decode(payload, data);
}

/* Must be called with iomutex held */
template<typename T> void Agilent4uhv::_write_window(const int win, const T &data)
{
	vector<unsigned char> payload, discard;
	encode(data, payload);
	_send_receive_message(win, WRITE_WINDOW, payload, discard);
}

template<typename T> void Agilent4uhv::write_window(const int win, const T &data)
{
	vector<unsigned char> payload, discard;
	encode(data, payload);
	send_receive_message(win, WRITE_WINDOW, payload, discard);
}

template<typename T> void Agilent4uhv::read_window_from_cache(const int win , T &data)
{
	vector<unsigned char> payload;
	cache.get(win, payload);
	assert(! payload.empty());
	decode(payload, data);
}
	
/* Must be called with iomutex held */
void Agilent4uhv::_read_window_into_cache(const int win)
{
	vector<unsigned char> discard, payload;
	send_receive_message(win, READ_WINDOW, discard, payload);
	cache.set(win, payload);
}

void Agilent4uhv::read_window_into_cache(const int win)
{
	vector<unsigned char> discard, payload;
	send_receive_message(win, READ_WINDOW, discard, payload);
	cache.set(win, payload);
}

void Agilent4uhv::check_state()
{
	int m;
	read_window_from_cache(8, m);
	string mode;
	switch (m) {
		case 0:	mode = "Serial"; break;
		case 1: mode = "Remote"; break;
		case 2:	mode = "Local";	break;
		case 3:	mode = "LAN"; break;
	};
	int status;
	read_window_from_cache(205, status);
	switch (status) {
		case 0:
			set_state(Tango::ON);
			set_status("The device is in ON state (" + mode + ")");
			break;
		case 111111:
			set_state(Tango::UNKNOWN);
			set_status("The device is in UNKNOWN state (" + mode + ")");
			break;
		default:
			set_state(Tango::FAULT);
			set_status("The device is in FAULT state (" + mode + ")");
	}
}

void* Agilent4uhvWorker::run_undetached(void * /*ptr*/)
{
	DEBUG_STREAM << "Worker::run_undetached() - enter" << endl;

	vector<int>::const_iterator it;
	it = device->polled_windows.begin();

	while(cont) {
		DEBUG_STREAM << "Worker::run_undetached() - loop" << endl;
		try {
			usleep(device->pollingSleep);
			if (*it > 1000) {
				/* Set channel */
				omni_mutex_lock guardio(device->iomutex);
				device->_write_window(505, *it - 2060);
				vector<unsigned char> discard, payload;
				/* Retrieve status of the previously set up channel */
				device->_send_receive_message(206, READ_WINDOW, discard, payload);
				/* Save as 206X into the cache */
				device->cache.set(*it, payload);
			} else {
				device->read_window_into_cache(*it);
			}
			if (++it == device->polled_windows.end()) {
				it = device->polled_windows.begin();
			}
		} catch(Tango::DevFailed &e) {
			/* Restart from begin */
			it = device->polled_windows.begin();
			/* Set UNKNOWN state */
			static const int m[] = {STX, ADDR, '2', '0', '5', READ_WINDOW,
				'1', '1', '1', '1', '1', '1', ETX, 0x38, 0x34};
			vector<unsigned char> message(m, m + sizeof(m) / sizeof(m[0]));
			device->cache.set(205, message);
			assert(false);
		}
	}

	DEBUG_STREAM << "Worker::run_undetached() - exit" << endl;

	return NULL;
}

/*----- PROTECTED REGION END -----*/	//	Agilent4uhv::namespace_ending
} //	namespace
